// This module is used by both the deamon and the client processes.
// It defines common datastructures and the contest for the shared memory
// area.

#include <netinet/ether.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <limits.h>
#include <math.h>
#include <assert.h>

// Used everywhere to save error messages returned from functions:
extern char errmsg[256];

//-----------------------------------------------------------------------------
// Generel non-applications specific structurs

// This represents a time in GMT format in seconds and microseconds sincd 1970.
struct Time {
  timeval tv;
  
  Time() { tv.tv_sec=0; tv.tv_usec=0; }

  int operator==(const Time& t) const { return tv.tv_sec==t.tv.tv_sec &&
                                               tv.tv_usec==t.tv.tv_usec; }
					       
  int operator<=(const Time& t) const { 
    return tv.tv_sec<t.tv.tv_sec || (tv.tv_sec==t.tv.tv_sec && tv.tv_usec<=t.tv.tv_usec);
  }
  
  // Returns the difference in seconds:
  double operator-(const Time& t) { 
    return (tv.tv_sec-t.tv.tv_sec)+(tv.tv_usec-t.tv.tv_usec)/(1000.0*1000.0);
  }  
  
  // Adds a number of seconds:  
  Time& operator+=(long s) {
    tv.tv_sec+=s;
    return *this;
  }    
  
  // Returns a string representation of the string:
  const char* tostr() const;
};

// This represent an unsigned integer consisting of two integers of type T
template<class T> struct DoubleInt {
  T ls; // Least significant part
  T ms; // Most signinificant part
 
  // Can by used by qsort, for example:
  static int compare(const void* a,const void* b) {
    if(*((DoubleInt*)a) <= *((DoubleInt*)b))
      if(*((DoubleInt*)a) == *((DoubleInt*)b))
        return 0;
      else
        return -1;
    else
      return 1;
  }

  DoubleInt(T a=0, T b=0) : ls(a), ms(b) {}

  DoubleInt operator+(const DoubleInt& a) const {
    return DoubleInt(T(ls+a.ls),T(ls+a.ls)<ls ? T(ms+a.ms+1) : T(ms+a.ms));
  }
  
  DoubleInt operator-(const DoubleInt& a) const {
    assert(a<=*this); // Assert that value doesn't get negative
    return DoubleInt(T(ls-a.ls),T(ls-a.ls)>ls ? T(ms-a.ms-1) : T(ms-a.ms));
  }

  DoubleInt operator+=(const DoubleInt& a) {
    (*this)=(*this)+a;
  } 
  
  DoubleInt operator-=(const DoubleInt& a) {
    (*this)=(*this)-a;
  }
  
  int IsZero() const { return ls==0 && ms==0; } 
  
  int operator<=(const DoubleInt& a) const {
    return (ms<a.ms || (ms==a.ms && ls<=a.ls));
  }

  int operator==(const DoubleInt& a) const {
    return (ms==a.ms && ls==a.ls);
  }
  
  T AsSingleInt() const { assert(!ms); return ls; }
  
  DoubleInt ShiftRight(int i) const { // Shifts i bits to the right
    const T_bits=(CHAR_BIT)*sizeof(T);
    assert(T_bits>=i);
    return DoubleInt((ls>>i)+((ms & ((1<<i)-1))<<(T_bits-i)), ms>>i);
  }  

  double AsDouble() const {
    // Womder how fast this one is..
    const T_bits=(CHAR_BIT)*sizeof(T);    
    return ((double)ls)+(ms ? ((double)ms)*pow(2,T_bits) : 0.0);
  }  
};

// This values is used for differnt counters. An unsigned is normally
// at least 32 bits so Int64 will be at least 64 bits which in most
// cases should be enough for the counters
typedef DoubleInt<unsigned> Int64;

//-----------------------------------------------------------------------------
// Generel application specific data structurs:

// A MAC address
class mac_addr {
  ether_addr ea;
  
  public:    
  mac_addr() {}
  mac_addr(const ether_addr& e) : ea(e) {}

  mac_addr& operator = (const mac_addr& m) { memcpy(ea.ether_addr_octet,m.ea.ether_addr_octet,sizeof(ea.ether_addr_octet)); return *this; }
  mac_addr& operator = (const ether_addr& m) { memcpy(ea.ether_addr_octet,m.ether_addr_octet,sizeof(ea.ether_addr_octet)); return *this; }
  mac_addr& operator = (u_int8_t m[ETH_ALEN]) { memcpy(ea.ether_addr_octet,m,sizeof(ea.ether_addr_octet)); return *this; }  

  operator const ether_addr*() const { return &ea; }
  int operator == (const mac_addr& m) const { return !memcmp(ea.ether_addr_octet,m.ea.ether_addr_octet,sizeof(ea.ether_addr_octet)); }
  int operator != (const mac_addr& m) const { return memcmp(ea.ether_addr_octet,m.ea.ether_addr_octet,sizeof(ea.ether_addr_octet)); }
  
  const char* tostr() const;
  static int getmaxlen_tostr(); // maximal length of returned value from tostr  
  int fromstr(const char*); // Returns 0 on succes  

  static int compare(const mac_addr& a, const mac_addr& b) { return a!=b; }  
  static int compare(const void* a, const void* b) { return compare(*((mac_addr*)a),*((mac_addr*)b)); }
  
  void setNull();
  int isBrdCast();
};

// An IP address:
struct ip_addr {
  in_addr ia;
  
  public:
  ip_addr() {}
  ip_addr(const in_addr& i) : ia(i) {}
  operator in_addr() const { return ia; }
  int operator == (const ip_addr& i) const { return ia.s_addr==i.ia.s_addr; }
  int operator != (const ip_addr& i) const { return ia.s_addr!=i.ia.s_addr; }
  ip_addr& operator = (const ip_addr& i) { ia.s_addr=i.ia.s_addr; return *this; }
  ip_addr& operator = (in_addr i) { ia.s_addr=i.s_addr; return *this; }
  ip_addr& operator = (unsigned int i) { ia.s_addr=i; return *this; }

  const char* tostr(int comp=4) const;    // comp=Number of components
  static int getmaxlen_tostr(int comp=4); // maximal length of returned value from tostr  
  int fromstr(const char*); // Returns 0 on succes

  static int compare(const ip_addr& a, const ip_addr& b) { return ntohl(a.ia.s_addr)-ntohl(b.ia.s_addr); }
  static int compare(const void* a, const void* b) { return compare(*((ip_addr*)a),*((ip_addr*)b)); }
  
  void setNull();
};

//-----------------------------------------------------------------------------
// Format of the shared memory area:

const int maxcounterc=32;
const int MaxCounterNameLen=10;

// Information about a single network card.
struct CardInfo {
  mac_addr macaddr;  // The MAC adress
  ip_addr  ipaddr;   // The last used IP adress
  Time updated;      // The last time the information for this card was updated
  
  Int64 Counters[maxcounterc]; // All the counters.  

  // Subtracts all the counters in ci from ourself. 
  // counterc should indicate the number of used counters.
  CardInfo& sub(CardInfo& ci,int counterc) {
    assert(macaddr==ci.macaddr);    
    for(int a=0; a<counterc; a++) {
      assert(ci.Counters[a]<=Counters[a]);
      Counters[a]-=ci.Counters[a];
    }
    return *this;
  }

  // Adds the counters:
  CardInfo& add(const CardInfo& ci,int counterc) {     
    for(int a=0; a<counterc; a++) {
      Counters[a]+=ci.Counters[a];
    }
    return *this;
  }

  // Are all the counters 0?
  int IsZero(int counterc) const {
    for(int a=0; a<counterc; a++) if(!Counters[a].IsZero()) return 0;
    return 1;
  }

  // These functions are used to do sorting:  
  static int reverse; // Reverse sorting? Should be left 0 after sorting
  
  static int rev(int a) { return reverse ? -a : a; }
  
  static int compare_mac(const void* a,const void* b) {
    return rev(mac_addr::compare(((CardInfo*)a)->macaddr,((CardInfo*)b)->macaddr));
  }
  
  static int compare_ip(const void* a,const void* b) {
    return rev(ip_addr::compare(((CardInfo*)a)->ipaddr,((CardInfo*)b)->ipaddr));
  }  

  static int counter; // Sort by the given counter
  static int compare_counter(const void* a,const void* b) {
    return rev(Int64::compare(&((CardInfo*)a)->Counters[counter],
                              &((CardInfo*)b)->Counters[counter]));
  }
};

// This is the shared memory area:
struct ServerMem {
  // Number of counters updated by this daemon.
  int Counterc;  
  
  // The names of the different counters:
  char CounterNames[maxcounterc][MaxCounterNameLen+1];
  
  Time start;    // The start time of the daemon.

  int maxcardc;  // Maximal number of cards to save information about.
  int curcardc;  // Number of cards we have currently.

  CardInfo cards[0]; // An array of maxcards elements with all our cards.
                     // These must be sorted by the MAC address
};

// Default filenames:
#define DEFAULT_LOGFILE    "/var/log/wipld"
#define DEFAULT_CNFFILE    "/etc/wipld.conf"
#define DEFAULT_PIDFILE    "/var/run/wipld.pid"
#define DEFAULT_DAEMONFILE "/etc/wipld.conf"
